#include "DualVector.h"

#include "Example.h"
#include "Timer.h"
#include "OptUtils.h"

#include <iostream>
#include <stdlib.h>
#include <stdio.h>
#include <assert.h>

#define myname "DualVector"

using namespace std;

namespace egstra {
  dual_vector::dual_vector(const int N, const string& dir,
			   const bool gzipped,
			   const string& prefix)
    :  _N(N), _cachedir(dir), _prefix(prefix), _theta(NULL),
       _gzipped(gzipped), _usedisk(dir.size() > 0),
       _rd_fname(NULL), _wr_fname(NULL),
       _rd_offset(-1), _wr_offset(-1) {
    if(!_usedisk) { /* memory cache */
      _theta = (const double**)calloc(_N, sizeof(double*));
    } else { /* disk cache */
      /* test for writing to the directory */
      const string fname(dir + "/tmp");
      FILE* const test = fopen(fname.c_str(), "wb");
      if(test == NULL) {
	cerr << myname << " : error writing to cache-dir \""
	     << dir << "\"" << endl;
	assert(0);
      }
      assert(fclose(test) == 0);
      assert(remove(fname.c_str()) == 0);
      /* set up the file-name templates */
      const int size = dir.size() + prefix.size() + 64;
      _rd_fname = (char*)malloc(size*sizeof(char));
      _wr_fname = (char*)malloc(size*sizeof(char));
      if(_gzipped) { /* gzipped binary cache */
	snprintf(_rd_fname, size, "gunzip -c \"%s/%s%08d.bin.gz\"",
		 _cachedir.c_str(), _prefix.c_str(), 0);
	/* offset includes gunzip -c ", dir, slash, prefix */
	_rd_offset = 11 + _cachedir.size() + 1 + _prefix.size();
	snprintf(_wr_fname, size, "gzip -c > \"%s/%s%08d.bin.gz\"",
		 _cachedir.c_str(), _prefix.c_str(), 0);
	/* offset includes gzip -c > ", dir, slash, prefix */
	_wr_offset = 11 + _cachedir.size() + 1 + _prefix.size();
      } else { /* normal binary cache */
	snprintf(_rd_fname, size, "%s/%s%08d.bin",
		 _cachedir.c_str(), _prefix.c_str(), 0);
	snprintf(_wr_fname, size, "%s/%s%08d.bin",
		 _cachedir.c_str(), _prefix.c_str(), 0);
	/* offset includes dir-name, slash, prefix */
	_rd_offset = _cachedir.size() + 1 + _prefix.size();
	_wr_offset = _rd_offset;
      }
    }
  }

  dual_vector::~dual_vector() {
    if(_cachedir.size() == 0) {
      for(int i = 0; i < _N; ++i) {
	if(_theta[i] != NULL) { delete [] _theta[i]; }
      }
      free(_theta);
    } else {
      free(_rd_fname);
      free(_wr_fname);
    }
  }


  const double* dual_vector::get(const pattern* const x) {
    const int xid = x->id();
    assert(xid >= 0);
    assert(xid < _N);
    if(!_usedisk) {
      assert(_theta[xid] != NULL);
      return _theta[xid];
    } else {
      /* write *only* the xid number */
      sprintf(_rd_fname + _rd_offset, "%08d", xid);
      _rd_fname[_rd_offset + 8] = '.';
      FILE* const f =
	(_gzipped ? popen(_rd_fname, "r") : fopen(_rd_fname, "rb"));
      if(f == NULL) {
	cerr << myname << " : error reading from \""
	     << _rd_fname << "\"" << endl;
	assert(0);
      }
      const int R = x->nparts();
      double* const tharr = new double[R];
      const int ntheta = fread(tharr, sizeof(double), R, f);
      assert(ntheta == R);
      /* check for end-of-file */
      int tmp;
      const int zero = fread(&tmp, sizeof(int), 1, f);
      assert(zero == 0);
      assert(feof(f));
      if(_gzipped) { assert(pclose(f) != -1); }
      else         { assert(fclose(f) == 0); }
      /* set return values */
      return tharr;
    }
  }

  void dual_vector::set(const pattern* const x,
			const double* const theta) {
    const int xid = x->id();
    assert(xid >= 0);
    assert(xid < _N);
    if(!_usedisk) {
      assert(theta != NULL);
      _theta[xid] = theta;
    } else {
      sprintf(_wr_fname + _wr_offset, "%08d", xid);
      _wr_fname[_wr_offset + 8] = '.';
      FILE* const f =
	(_gzipped ? popen(_wr_fname, "w") : fopen(_wr_fname, "wb"));
      if(f == NULL) {
	cerr << myname << " : error writing to \""
	     << _wr_fname << "\"" << endl;
	assert(0);
      }
      const int R = x->nparts();
      const int ntheta = fwrite(theta, sizeof(double), R, f);
      assert(ntheta == R);
      if(_gzipped) { assert(pclose(f) != -1); }
      else         { assert(fclose(f) == 0); }
      /* discard the arrays after writing them to disk */
      delete [] theta;
    }
  }

  void dual_vector::discard(const double* const theta) {
    /* if we are not using disk, there is nothing to do in this
       function */
    if(_usedisk) { delete [] theta; }
  }
}

#undef myname
