/************************************************************
 *
 *   Author : Terry Koo
 *            maestro@csail.mit.edu
 *
 *   Objective functions for optimization
 *
 ************************************************************/
#ifndef EGSTRA_OBJECTIVEFUNCTION_H
#define EGSTRA_OBJECTIVEFUNCTION_H

#include "Example.h"
#include "FGen.h"
#include "StructuredModel.h"
#include "Parameters.h"

namespace egstra {


  /* tolerance for violation of the primal objective constraints
     log-prob <= 0  or  hinge-loss >= 0 */
#define EGSTRA_OBJECTIVEFUNCTION_PRIMAL_TOL 1e-8

  /* maximize the conditional log-likelihood of a globally-normalized
     log-linear distribution */
  class obj_loglinear {
  public:

    static const char* name;

    /* theta' = (1 - scale)*theta + scale*(W . f(x,part)) */
    static double update_dual_variable(const double theta,
				       const double iprod,
				       const double loss,
				       const double eta) {
      return theta + eta * (iprod - theta);
    }

    /* the dual is the negative entropy:

       -H = marg . theta - logZ

       thus, the initializer is -logZ and the per-part contribution is
       marg(part)*theta(part) */
    static double exdual_init(const double logZ) { return -logZ; }
    static double exdual_contrib(const double mu,
				 const double theta,
				 const double loss) {
      return mu * theta;
    }

    /* return the log-likelihood of this sentence. */
    static double exprimal(const example* const ex,
			   double* const S,
			   structured_model& M);

    /* compute the scaled contribution of x to the gradient of the
       log-likelihood, and add it to g.  also returns the scaled
       log-likelihood. */
    static double exgrad(parameters& g,
			 const double scale,
			 const example* const ex,
			 double* const S,
			 const struct fvec* const F,
			 structured_model& M);

    /* for log-linear estimation, a learning rate higher than one is
       not sensible */
    static double lrate_search_init() { return 1.0; }

    /* whether this objective function has a sensible probabilistic
       interpretation */
    static bool probabilistic() { return true; }
  };




  /* maximize the loss-adjusted margin of a hyperplane classifier */
  class obj_maxmargin {
  public:
    static const char* name;

    /* theta' = theta + scale*(W . f(x,part) + loss(part)) */
    static double update_dual_variable(const double theta,
				       const double iprod,
				       const double loss,
				       const double eta) {
      return theta + eta * (iprod + loss);
    }

    /* the dual (for minimization) is the negative expected loss

       -E[L] = -marg . loss

       thus the initializer is 0 and the per-part contribution is
       -marg(part)*loss(part) */
    static double exdual_init(const double logZ) { return 0; }
    static double exdual_contrib(const double mu,
				 const double theta,
				 const double loss) {
      return -(mu * loss);
    }

    /* return the negative hinge loss.  NB: S will be modified. */
    static double exprimal(const example* const ex,
			   double* const S,
			   structured_model& M);

    /* compute the scaled contribution of x to the subgradient of the
       (negative) hinge loss, and add it to g.  also returns the
       scaled negative hinge loss.  NB: S will be modified. */
    static double exgrad(parameters& g,
			 const double scale,
			 const example* const ex,
			 double* const S,
			 const struct fvec* const F,
			 structured_model& M);

    /* for max-margin learning, high learning rates may be possible */
    static double lrate_search_init() { return 1024.0; }

    /* whether this objective function has a sensible probabilistic
       interpretation */
    static bool probabilistic() { return false; }
  };


}

#endif /* EGSTRA_OBJECTIVEFUNCTION_H */
