#include <stdio.h>
#include <assert.h>
#include <iostream>

#include "Parameters.h"

#define myname "Parameters"

using namespace std;

namespace egstra {

  parameters::parameters(const int dim)
    : _dim(dim), _W(NULL), _alpha(1.0), _Wsq(0), _self_alloc(true)
  {
    realloc(_dim);
  }

  parameters::parameters(double* const V, const int dim)
    : _dim(dim), _W(V), _alpha(1.0), _Wsq(0), _self_alloc(false)
      /* NB: _self_alloc = false prevents deletion of _W */
  {
    assert(_dim > 0);
    realloc(_dim);
    /* set the proper squared-norm */
    assert(_alpha == 1.0);
    _Wsq = 0;
    for(int i = 0; i < _dim; ++i) {
      const double val = _W[i];
      _Wsq += val*val;
    }
  }

  parameters::~parameters() { dealloc(); }




  void parameters::dealloc() {
    if(_W != NULL && _self_alloc) { delete [] _W; }
  }

  void parameters::realloc(const int dim) {
    dealloc();
    _dim = dim;
    assert(_dim >= 0);

    if(dim > 0) {
      if(_self_alloc) { /* create a zeroed vector */
	_W = new double[_dim]; zero();
      }
      assert(_W != NULL);
    } else {
      _W = NULL;
    }
  }


  void parameters::save(const string& dir, const int t,
			const string& stem) const {
    char* const fname = new char[64 + stem.size()];
    sprintf(fname, "%s.%03d.gz", stem.c_str(), t);
    char* const command = new char[64 + stem.size() + dir.size()];
    sprintf(command, "gzip -c > '%s/%s'", dir.c_str(), fname);

    cerr << myname << " : writing to \""
         << dir << "/" << fname << "\" " << flush;

    FILE* gzout = popen(command, "w");
    assert(gzout != NULL);
    fprintf(gzout, "%d\n", _dim);

    delete [] fname;
    delete [] command;

    /* only print nonzero parameters */
    int nnz = 0;
    if(_alpha != 0) {
      for (int i = 0; i < _dim; ++i) {
	if(_W[i] != 0) {
	  fprintf(gzout, "%d %.16g\n", i, _alpha*_W[i]);
	  if(((++nnz) & 0xfffff) == 0) { cerr << "." << flush; }
	}
      }
    }

    assert(pclose(gzout) != -1);
    cerr << " [nnz=" << nnz << "/" << _dim << "]" << endl;
  }

  void parameters::load(const string& dir, const int t,
			const string& stem) {
    assert(_self_alloc);
    char* const fname = new char[64 + stem.size()];
    sprintf(fname, "%s.%03d.gz", stem.c_str(), t);
    char* const command = new char[64 + stem.size() + dir.size()];
    sprintf(command, "gunzip -c '%s/%s'", dir.c_str(), fname);

    cerr << myname << " : reading from \""
         << dir << "/" << fname << "\" " << flush;

    FILE* gzin = popen(command, "r");
    assert(gzin != NULL);
    _dim = -1;
    assert(fscanf(gzin, "%d\n", &_dim) == 1);
    assert(_dim > 0);
    realloc(_dim);

    delete [] fname;
    delete [] command;

    /* the parameters file can omit zero values, so we need to read
       lines on-demand */
    int index = -1;
    double value = 0;
    int nnz = 0;
    while(1) {
      const int nread = fscanf(gzin, "%d %lf\n", &index, &value);
      if(nread <= 0 && feof(gzin)) { break; }
      assert(nread == 2);
      assert(index >= 0);
      assert(index < _dim);
      _W[index] = value;
      if(((++nnz) & 0xfffff) == 0) { cerr << "." << flush; }
    }

    _alpha = 1.0;
    _Wsq = 0;
    for(int i = 0; i < _dim; ++i) {
      const double val = _W[i];
      _Wsq += val*val;
    }

    assert(pclose(gzin) != -1);
    cerr << " [nnz=" << nnz << "/" << _dim << "]" << endl;
  }


  parameters* parameters::copy() const {
    parameters* ret = new parameters(_dim);
    double* const retbuf = ret->_W;
    for(int i = 0; i < _dim; ++i) { retbuf[i] = _W[i]; }
    ret->_alpha = _alpha;
    ret->_Wsq = _Wsq;
    return ret;
  }

}

#undef myname
