/**
 * EG for Structured Algorithms
 * @file egstra.cc
 * @author Xavier Carreras, Terry Koo
 */

#include <iostream>
#include <string>

#include <stdlib.h>

#include "Constants.h"
#include "Options.h"
#include "OptUtils.h"
#include "StructuredModel.h"
#include "FGen.h"
#include "Trainer.h"
#include "Registry.h"

using namespace std;
using namespace egstra;

static void usage(const char* mesg = "") {
  cerr << "egstra " << VERSION << " --- {carreras,maestro}@csail.mit.edu" << endl;
  cerr << endl;
  cerr << "egstra options:" << endl;
  cerr << " --dir=<pathname>     : output directory for various files" << endl;
  cerr << " --T=<int>            : number of training rounds" << endl;
  cerr << " --trdata=<pathname>  : training data file" << endl;
  cerr << " --valdata=<pathname> : held-out data file, to be classified" << endl;
  cerr << "                        on each training iteration" << endl;
  cerr << endl << "egstra modules: " << endl;
  cerr << " --opt=<string>       : type of optimization" << endl;
  /* print the list of available trainers */
  cerr << "      available:";
  vector<string> names;
  registry::getnames("trainer", names);
  for(int i = 0; i < (int)names.size(); ++i) { cerr << " " << names[i]; }
  names.clear();
  cerr << endl << endl;
  cerr << " --fgen=<string>      : name of feature generator" << endl;
  /* print the list of available trainers */
  cerr << "      available:";
  registry::getnames("fgen", names);
  for(int i = 0; i < (int)names.size(); ++i) { cerr << " " << names[i]; }
  names.clear();
  cerr << endl << endl;
  cerr << " --model=<string>     : name of structured model" << endl;
  /* print the list of available trainers */
  cerr << "      available:";
  registry::getnames("model", names);
  for(int i = 0; i < (int)names.size(); ++i) { cerr << " " << names[i]; }
  names.clear();
  cerr << endl << endl;
  cerr << mesg << endl;
}

static void help() {
  usage();

  string opt_name;
  if(options::get("opt", opt_name)) {
    trainer* const TR = (trainer*)
      registry::construct("trainer", opt_name.c_str());
    TR->usage();
    delete TR;
  }

  string model_name;
  if(options::get("model", model_name)) {
    structured_model* const M = (structured_model*)
      registry::construct("model", model_name.c_str());
    M->usage();
    delete M;
  }

  string fgen_name;
  if(options::get("fgen", fgen_name)) {
    fgen* const FG = (fgen*)
      registry::construct("fgen", fgen_name.c_str());
    FG->usage();
    delete FG;
  }

  exit(1);
}


int main(int argc, char ** argv)
{
  if(argc == 1) { usage(); exit(1); }

  /* any non-options (not beginning with '-') are assumed to be
     configuration files and are read */
  for(int i = 1; i < argc; ++i) {
    const char* const arg = argv[i];
    if(arg[0] != '-') {
      options::read(string(argv[1]));
    }
  }
  /* now parse the command-line arguments (potentially overwriting
     anything set in a config file) */
  options::read(argc, argv);

  cerr << "egstra " << VERSION << " called as:" << endl;
  for(int i = 0; i < argc; ++i) {
    cerr << (i == 0 ? "" : "    ")
	 << argv[i]
	 << (i == argc - 1 ? "" : " \\")
	 << endl;
  }

  /* check the overall options */

  int unused;
  if(options::get("help", unused) || options::get("h", unused)) {
    help();
  }

  int T;
  if(!options::get("T", T)) {
    usage("please provide a max number of iterations"); exit(1);
  }

  string tr_name;
  if(!options::get("trdata", tr_name)) {
    usage("please provide a training data file"); exit(1);
  }
  const char* valfile = NULL;
  string val_name;
  if(options::get("valdata", val_name)) {
      valfile = val_name.c_str();
  }

  string dir;
  if(!options::get("dir", dir)) {
    usage("please provide an output directory"); exit(1);
  }
  /* write options to a config file */
  options::write(dir + "/options");

  string opt_name;
  if(!options::get("opt", opt_name)) {
    usage("please name an optimization method"); exit(1);
  }
  trainer* const TR = (trainer*)
    registry::construct("trainer", opt_name.c_str());

  string model_name;
  if(!options::get("model", model_name)) {
    usage("please name a structured model"); exit(1);
  }
  structured_model* const M = (structured_model*)
    registry::construct("model", model_name.c_str());

  string fgen_name;
  if(!options::get("fgen", fgen_name)) {
    usage("please name a feature generator"); exit(1);
  }
  fgen* const FG = (fgen*)
    registry::construct("fgen", fgen_name.c_str());

  /* make sure to process the feature-generator's options last, as it
     will typically involve a dictionary-load */
  TR->process_options();
  M->process_options();
  FG->process_options();

  cerr << "egstra : feature dimensionality " << FG->dim() << endl;

  dataset trdata;
  dataset valdata;
  optutils::load_datasets(*M, tr_name.c_str(), trdata, valfile, valdata);

  TR->set_dir(dir);
  TR->set_model(M);
  TR->set_fgen(FG);
  TR->train(trdata, valdata, T);

  optutils::clear_dataset(trdata);
  optutils::clear_dataset(valdata);

  delete M;
  delete FG;
  delete TR;
}
